import json
import time
import asyncio
import logging
import subprocess

from typing import Optional, List, Any
from usody_sanitize import schemas

logger = logging.getLogger(__name__)


async def list_blocks(
        dev_path: Optional[str] = None,
        children: bool = False,
) -> List[schemas.Block]:
    """Runs the command `lsblk` to get the information about devices.

    Used maily to detect USB, live USB or disk attached.

    :param Optional[str] dev_path:
    :param bool children: Get partitions of each device.

    :return:
    """
    command = ["lsblk", "-JOa" if children else "-JOad"]
    logger.debug(f"Running command: `{command}`")
    if dev_path:
        command += [dev_path]
    output_text = subprocess.check_output(command, text=True)

    logger.debug(f"Processing `{command}` output.")
    devices_json = json.loads(output_text.strip()).get("blockdevices", [])
    return [schemas.Block.parse_obj(dev) for dev in devices_json]


async def get_smart_info(dev_path: str) -> schemas.Smart:
    """Gets the information of the device with the `smartctl` command.

    Used to detect when a disk is SSD or HDD correctly, as lsblk don't
    detect this value correctly.

    :param str dev_path:
    :return:
    """
    command = ["smartctl", "-aj", dev_path]
    proc = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await proc.wait()
    output = await proc.stdout.read()
    smart_json = json.loads(output.decode('UTF-8').rstrip())
    logger.debug(f"Smart for {dev_path} is: {smart_json}")
    return schemas.Smart.parse_obj(smart_json)


async def get_total_sectors(dev_path: str) -> int:
    """Get the total
    """
    command = ["blockdev", "--getsz", dev_path]
    proc = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    await proc.wait()
    if proc.returncode:
        raise Exception(f"Can not get total disk sectors from {dev_path}")
    return int(await proc.stdout.read())


async def erasure_command(
        command: str,
        process_manager: Optional[Any] = None,
) -> schemas.ErasureCommand:
    """Runs the command given, but it returns a `schemas.ErasureCommand`
    object with the processed data.

    :param List[str] command: Command string to be executed like a shell
        on the system.
    :param process_manager: Async function to allow to manipulate the
        command process while still running.

    :return:
    """
    if isinstance(command, list):
        command = ' '.join(command)

    cmd = schemas.ErasureCommand(command=command, start_time=time.time())
    proc = await asyncio.create_subprocess_shell(
        # "timeout 30 " +
        cmd.command,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    if process_manager:
        await process_manager(cmd, proc)

    # await proc.communicate()
    await proc.wait()
    cmd.end_time = time.time()

    if cmd.stdout is None:
        stdout = await proc.stdout.read()
        cmd.stdout = stdout.decode('UTF-8').rstrip()

    if cmd.stderr is None:
        stderr = await proc.stderr.read()
        cmd.stderr = stderr.decode('UTF-8').rstrip()

    cmd.return_code = proc.returncode
    cmd.success = proc.returncode == 0

    return cmd


async def read_from_sector(
        path_to_dev: str,
        sector: int,
        bs: int = 512,
) -> schemas.ErasureCommand:
    """Read the X bytes (bs) from a disk sector.

    :param str path_to_dev: Path to the device. Example: `/dev/sda`
    :param int sector: Sector to read on the disk.
    :param int bs: Sector sizes, by default 512 bytes.

    :return: schemas.ErasureCommand
    """
    read_command = f"dd if={path_to_dev} bs={bs} count=1 skip={sector}" \
                   " | xxd -ps"
    logger.debug(f"{path_to_dev}: Read data on sector {sector}.")
    return await erasure_command(read_command)


async def write_to_sector(
        path_to_dev: str,
        sector: int,
        bs: int = 512,
        zeros: bool = False,
) -> schemas.ErasureCommand:
    """Write X bytes (bs size must be provided) to the disk sector.

    :param str path_to_dev: Path to the device. Example: `/dev/sda`
    :param int sector: Sector to read.
    :param int bs: Sector sizes, by default 512 bytes.
    :param bool zeros: The pattern desired to use.

    :return: schemas.ErasureCommand
    """
    write_command = f"dd if={'/dev/zero' if zeros else '/dev/random'}"\
                    f" of=/dev/sda bs={bs} count=1 seek={sector}"
    logger.debug(
        f"Writing {'zeros' if zeros else 'random'} data into {path_to_dev}.")
    return await erasure_command(write_command)



import sys
import json
import logging
import asyncio

from enum import Enum

from typing import List, Union, Optional
from usody_sanitize import schemas, commands, steps, utils, __version__
from usody_sanitize.methods import (
    BASIC,
    BASELINE,
    ENHANCED,
    CRYPTOGRAPHIC,
)

logger = logging.getLogger(__name__)


class DefaultMethods(Enum):
    """To easily convert strings to the method pre-defined. Used
    when calling to the sanitize with string instead with method class.
    """
    BASIC = BASIC
    BASELINE = BASELINE
    ENHANCED = ENHANCED


async def auto_erase_disks(
        method: Optional[str] = None,
        disks: Optional[List[str]] = None,
) -> Optional[List[dict]]:
    """A main point to erase all disks unless disks mounted at root "/"
    then generates a certificate as JSON with all the information of
    the disk, erasure and the validation of the successful data wiped.

    :param Optional[str] method: A string to select which pre-defined
        methods to use, basic, baseline or enhanced. If none, Basic will
        be selected.
    :param Optional[List[str]] disks: A list of disks to be erased. If
        none, all disks detected will be erased and sanitized.

    :return:
    """
    # Select method.
    if method:
        try:
            method = DefaultMethods[method.upper()].value
        except ValueError:
            sys.exit("Sanitize method not valid.")
    else:
        # Default method when not set by args.
        method = DefaultMethods.BASIC.value
    logger.info(f"Using sanitize method '{method.name}'.")

    # Get disks.
    blocks: List[schemas.Block] = await commands.list_blocks(children=True)
    erasures: List[ErasureProcess] = []
    tasks: List[asyncio.Task] = []

    for blk in blocks:

        if blk.type != "disk":
            logger.debug(f"Skipping {blk.path}")
            continue  # Skip non disk volumes.

        if disks is not None:
            if blk.path in disks:
                disks.remove(blk.path)
            else:
                continue

        # Detect if one of those drives has a mounted partition as root.
        if blk.children:

            if any([cld.mountpoint == "/" for cld in blk.children
                    if cld.mountpoint is not None]):
                logger.warning(f"Skipping disk {blk.path} mounted as root.")
                continue

        logger.debug(f"Processing disk {blk.path}.")
        erasure = ErasureProcess(blk, method)
        erasures.append(erasure)

        # Start erasure task.
        tasks.append(asyncio.create_task(erasure.run()))

    if disks:
        logger.warning(f"Disks [{', '.join(disks)}] not found.")

    [await task for task in tasks]
    logger.debug(f">>> {erasures}")
    return [r.export() for r in erasures]


class ErasureProcess:
    def __init__(
            self,
            block: schemas.Block,
            method: Union[schemas.ErasureMethod, str] = None
    ):
        if isinstance(method, str):
            # Todo: convert to schema.
            pass

        logger.info(f"Selected device {block.path} for sanitization.")
        self._device = schemas.Device(
            export_data=schemas.ExportData(
                block=block,  # Data from `lsblk`
            )
        )
        validation = schemas.ErasureValidation()
        self._certificate = schemas.ErasureCertificate(
            device_info=self._device,
            validation=validation,
            usody_sanitize=__version__
        )
        # --> HERE SELECT THE DEFAULT ERASURE METHOD <--
        self._certificate.method = BASIC if method is None else method

    @property
    def blk(self) -> Optional[schemas.Block]:
        return self._device.export_data.block

    @property
    def smart(self) -> Optional[schemas.Smart]:
        return self._device.export_data.smart

    def export(self) -> dict:
        return self._certificate.dict()

    async def extract_device_info(self):
        """Get the basic info from the `lsblk` command and store it
        into the device schema. As the name of the function says,

        :return:
        """
        # Check if the device is SSD or HDD.
        if self.smart.rotation_rate == 0:  # Is SSD.
            if self.blk.rota == 1:  # `lsblk` says is a HDD.
                logger.debug("SSD \n\t"
                             f"`lsblk` rotate = {self.blk.rota}\n\t"
                             "`smartctl` rotation_rate"
                             f" = {self.smart.rotation_rate}")

            self._device.storage_medium = "SSD"

        else:
            self._device.storage_medium = "HDD"

        logger.debug(f"Device {self.blk.path} has been validated.")

    async def run(self):
        # Do `smartctl` here
        self._device.export_data.smart = await commands \
            .get_smart_info(self.blk.path)

        # Extract device info from `lsblk` and `smartctl`.
        await self.extract_device_info()

        logger.debug(f"Erasure process for {self.blk.path} started.")

        if self._certificate.method.verification_enabled:
            # Pre validation steps before erasure.
            await self._pre_validation()

        if self._certificate.device_info.storage_medium == 'HDD':
            # If rotates, means it has magnetic disks.
            logger.info(f"Device {self.blk.path} is a HDD")
            logger.debug(f"Data of the device: {self.blk}")
            await self._erase_hdd()

        elif self._certificate.device_info.storage_medium == 'SSD':
            logger.info(f"Device {self.blk.path} is an SSD")
            logger.debug(f"Data of the device: {self.blk}")
            await self._erase_ssd()
        else:
            # Todo: Research about SSDHD types.
            raise Exception("Unknown method for SSDHD")

        if self._certificate.method.verification_enabled:
            # Validation step after erasure.
            await self._validation()

        # The result depends on the validation
        if self._certificate.method.verification_enabled:
            self._certificate.result = self._certificate.validation.result
        elif self._certificate.erasure_steps:
            # IF validation is disabled, just check the erase command.
            self._certificate.result = self._certificate \
                .erasure_steps[-1].success
        else:
            # If there is no steps and neither validation,
            # there is no erasure.
            self._certificate.result = False

        # Show info when the erasure is done.
        logger.debug(f"{self.blk.path}: Erasure finished, results:"
                     f" {json.dumps(self._certificate.dict(), indent=4)}")

    async def _pre_validation(self):
        """Check if the disk is not mounted and if it is not a
        read-only device.

        :return:
        """
        bs = self.smart.logical_block_size
        max_sectors = await commands.get_total_sectors(self.blk.path)
        sectors = utils.get_spaced_numbers(max_sectors, 10)

        def _successful_command(
                _cmd: schemas.ErasureCommand,
        ):
            self._certificate.validation.commands.append(_cmd)
            if _cmd.return_code != 0:
                _cmd.success = False
                # Todo: Clean only the data from sectors.
                self._certificate.validation.data = {}
                logger.warning(f"Validation step {_cmd.command} failed.")
                return False
            return True

        for s in sectors:
            # First command.
            cmd1 = await commands.read_from_sector(self.blk.path, s, bs)
            cmd1.description = f"Read data from sector {s} to validate" \
                               " if have been changed."
            if not _successful_command(cmd1):
                return False
            self._certificate.validation.data[s] = cmd1.stdout
            cmd1.stdout = "Private"

        for s in sectors:
            # Second command.
            cmd2 = await commands.write_to_sector(self.blk.path, s, bs)
            cmd2.description = "Write the data to validate into the sectors"
            if not _successful_command(cmd2):
                return False

        for s in sectors:
            # Third command.
            cmd3 = await commands.read_from_sector(self.blk.path, s, bs)
            cmd3.description = "Check if new bytes has been written"
            if not _successful_command(cmd3):
                return False

            elif self._certificate.validation.data[s] == cmd3.stdout:
                logger.warning(
                    f"{self.blk.path}: "
                    f"Validation failed: Sector {s} has not been changed")
                return False
            # Successful write, update the sector with new value.
            self._certificate.validation.data[s] = cmd3.stdout

        logger.debug(f"{self.blk.path}: Pre validation step finished.")
        return True

    async def _validation(self):
        """
        :return:
        """
        for sector in self._certificate.validation.data:
            cmd = await commands.read_from_sector(
                self.blk.path, sector, self.smart.logical_block_size)

            if cmd.stdout == self._certificate.validation.data[sector]:
                self._certificate.validation.result = False
                logger.warning(f"{self.blk.path}: Erasure validation failed.")
                return

        self._certificate.validation.result = True
        logger.debug(f"Validation passed.")

    async def _erase_hdd(self):
        """Reads the erasure method name for the erasure and calls the
        right method with the right arguments to automatically erase
        the disk as defined.
        :return:
        """
        # @TODO: create a pre step for validation, first format
        #   the drive and create some content to be deleted latter.
        #   OR use dd to write a few bytes to the disk and check if
        #   they are still there after the erasure.

        current_method = self._certificate.method.program

        logger.debug(f"Using '{current_method}' for `{self.blk.path}`.")
        # Todo: Run the erasure as many steps as defined in the method.
        for n in range(self._certificate.method.overwriting_steps):
            # Erasure steps.
            if current_method == 'shred':
                step = await steps.erase_hdd_shred(self.blk.path)
                self._certificate.erasure_steps.append(step)
            elif current_method == 'badblocks':
                step = await steps.erase_hdd_badblocks(self.blk.path)
                self._certificate.erasure_steps.append(step)
            else:
                logger.error(
                    f"'{self._certificate.method.program}' not implemented.")

    async def _erase_ssd(self):
        """To erase an SSD, the only method allowed right now is the
        Baseline Cryptographic, so the method will be overwritten.
        :return:
        """
        # SSD can only be erased with method Cryptographic. Overwriting
        # erasures damages the disk.
        self._certificate.method = CRYPTOGRAPHIC

        # Create the step schema and set initial values.
        logger.debug(f"Starting erasure step for {self.blk} as SSD.")
        self._certificate.erasure_steps.append(
            await steps.erase_ssd_hdparm(self.blk.path),
        )



from .export_data import Block, Smart, ExportData
from .devices import Device
from .erasure import (
    ErasureCertificate,
    ErasureValidation,
    ErasureCommand,
    ErasureMethod,
    ErasureStep,
)



from typing import Optional, List
from pydantic import BaseModel, Field

from usody_sanitize.schemas.devices import Device
from usody_sanitize import __version__ as app_version


class ErasureCommand(BaseModel):
    """Main and base class to define a collection of steps to proceed.
    """
    description: Optional[str] = Field(
        default=None, description="Step description")
    command: str = Field(default=..., description="Command used on this step")

    stdout: Optional[str] = Field(
        default=None, description="normal output from the command executed")
    stderr: Optional[str] = Field(
        default=None, description="error output from the command executed")

    return_code: Optional[int] = Field(
        default=None, description="Command return code")

    success: bool = Field(
        default=False, description="Tells if the step has been executed"
                                   " correctly")

    start_time: float = Field(
        default=..., description="Exact time when the command started")
    end_time: Optional[float] = Field(
        default=None, description="Exact time when the command ended")


class ErasureStep(BaseModel):
    """Main and base class to define a collection of steps to proceed.
    """
    step: int = Field(default=..., description="Step number")
    date_init: Optional[float] = Field(
        default=None, description="Date at start of the step")
    date_end: Optional[float] = Field(
        default=None, description="Date at the end of the step")
    duration: Optional[float] = Field(
        default=None, description="Duration of the step")
    commands: List[ErasureCommand] = Field(
        default=[], description="")
    success: bool = Field(default=False, description="Tells if the step has"
                                                     "been executed correctly")


class ErasureValidation(BaseModel):
    """Defines the validation process result.
    """
    result: Optional[bool] = Field(
        default=None, description="Defines if the validation is successful")
    commands: List[ErasureCommand] = Field(
        default=[], description="A list of commands used to "
                                "validate this process")
    data: dict = Field(
        default={}, description="Bytes read from each disk sector")


class ErasureMethod(BaseModel):
    name: str = Field(
        default=..., description="Erasure method name (Basic,"
                                 " Baseline Erasure, Baseline Cryptographic"
                                 " or Enhanced)")
    standard: str = Field(
        default=..., description="Standard name (NIST SP-800-88 /  HMG"
                                 " Infosec Standard 5 Baseline Standard /"
                                 " HMG Infosec Standard 5 Baseline Enhanced")
    description: Optional[str] = Field(
        default=None, description="Description of the method")
    removal_process: Optional[str] = Field(
        default=None, description="Demagnetization / Destruction /"
                                  " Overwriting / Cryptographic erase")
    program: Optional[str] = Field(
        default=None, description="None / shred / badblocks / hdparm")
    verification_enabled: bool = Field(
        default=False, description="Will check if the erasure has deleted"
                                   " all the data on the disk by comparing"
                                   " some data written on the disk before"
                                   " the erasure with the data read after.")
    bad_sectors_enabled: bool = Field(
        default=False, description="If true, bad sectors will be checked"
                                   "on the erasure process")
    # sectors_total: Optional[int] = Field(
    #     default=None, description="Total number of sectors")
    # sectors_bad: Optional[int] = Field(
    #     default=None, description="Total of bad sectors")
    warnings: Optional[str] = Field(
        default=None, description="Summary of the check (no / Bad sectors)")
    overwriting_steps: Optional[int] = Field(
        default=None, description="Number of overwriting steps")

    class Config:
        fields = {
            "name": {"readonly": True},
        }


class ErasureCertificate(BaseModel):
    erasure_steps: List[ErasureStep] = Field(
        default=[], description="A list of erasure steps, each steps contain"
                                " a list of commands and the result of each"
                                " command.")
    validation: Optional[ErasureValidation] = Field(
        default=None, description="Information to certificate"
                                  "the validation process.")

    # Storage Drive
    device_info: Device = Field(
        default=..., description="all info about the device")

    # Method
    method: Optional[ErasureMethod] = Field(
        default=None, description="Erasure method")

    result: bool = Field(
        default=False, description="True means erasure has been pass"
                                   " correctly, False means something"
                                   " failed and the data could not be"
                                   " erased/wipe.")

    sanitize_version: str = Field(
        default=app_version,
        description="Version of usody_sanitize python package.")





"""Pre-defined sanitize methods to erase disks securely."""
from usody_sanitize import schemas


BASIC = schemas.ErasureMethod(
    name="Basic Erasure",
    # https://tsapps.nist.gov/publication/get_pdf.cfm?pub_id=917935
    standard="",
    description="A single-pass overwrite of the entire drive with"
                " zeros. This method is relatively fast and simple,"
                " but it may not be completely effective in destroying"
                " all traces of the original data.",
    removal_process="Overwriting",
    program="shred",
    verification_enabled=True,
    bad_sectors_enabled=False,
    overwriting_steps=1,
)

BASELINE = schemas.ErasureMethod(
    name="Baseline Erasure",
    standard="NIST, Infosec HGM Baseline",
    description="Method for securely erasing data in compliance"
                " with HMG Infosec Standard 5 guidelines includes"
                " a single step of a random write process on the"
                " full disk. This process overwrites all data with"
                " a randomized pattern, ensuring that it cannot be"
                " recovered. Built-in validation confirms that the"
                " data has been written correctly, and a final"
                " validation confirms that all data has been deleted.",
    removal_process="Overwriting",
    program="badblocks",
    verification_enabled=False,
    bad_sectors_enabled=True,
    overwriting_steps=1,
)

CRYPTOGRAPHIC = schemas.ErasureMethod(
    name="Baseline Cryptographic",
    standard="NIST, Infosec HGM Baseline",
    description="Method for securely erasing data in compliance"
                " with HMG Infosec Standard 5 guidelines includes"
                " a single step of a random write process on the"
                " full disk. This process overwrites all data with"
                " a randomized pattern, ensuring that it cannot be"
                " recovered. Built-in validation confirms that the"
                " data has been written correctly, and a final"
                " validation confirms that all data has been deleted.",
    removal_process="Overwriting",
    program="hdparm",
    verification_enabled=False,
)

ENHANCED = schemas.ErasureMethod(
    name="Enhanced Erasure",
    standard="HMG Infosec Standard 5",
    description="Method for securely erasing data in compliance"
                " with HMG Infosec Standard 5 guidelines includes"
                " a single step of a random write process on the"
                " full disk. This process overwrites all data with"
                " a randomized pattern, ensuring that it cannot be"
                " recovered. Built-in validation confirms that the"
                " data has been written correctly, and a final"
                " validation confirms that all data has been deleted.",
    removal_process="Overwriting",
    program="badblocks",
    verification_enabled=True,
    bad_sectors_enabled=True,
    overwriting_steps=3,
)
